@@ -1,9 +1,11 @@
 r"""
 ``platform`` 模块为PyWebIO提供了对不同Web框架的支持。
 
+具体用法参见用户手册 :ref:`与Web框架集成 <integration_web_framework>` 小节
+
 Tornado相关
 --------------
-.. autofunction:: start_server
+.. autofunction:: pywebio.platform.tornado.start_server
 .. autofunction:: pywebio.platform.tornado.webio_handler
 
 Flask相关
